"""tapdance.syncs - Module containing the sync() function and sync helper functions."""

import os
from typing import List, Optional

import uio
import runnow
from logless import get_logger, logged
from tapdance import config, docker, plans, states

logging = get_logger("tapdance")


def sync(
    tap_name: str,
    target_name: str = "csv",
    table_name: str = "*",
    taps_dir: Optional[str] = None,
    *,
    dockerized: Optional[bool] = None,
    rescan: bool = False,
    tap_exe: Optional[str] = None,
    target_exe: Optional[str] = None,
    config_dir: Optional[str] = None,
    config_file: Optional[str] = None,
    catalog_dir: Optional[str] = None,
    target_config_file: Optional[str] = None,
    state_file: Optional[str] = None,
    exclude_tables: Optional[List[str]] = None,
) -> None:
    """
    Synchronize data from tap to target.

    Parameters:
    ----------
    tap_name : {str}
        The name/alias of the source tap, without the `tap-` prefix.
    target_name : {str}
        The name/alias of the target, without the `tap-` prefix.
        (Default="csv")
    table_name : {str}
        The name of the table to sync. To sync multiple tables, specify
        a comma-separated list of tables surrounded by queare brackets (e.g. "[tb1,tbl2]"),
        or use "*" to sync all table.
        (Default="*")
    dockerized : {bool}
        True or False to force whether the command is run
        dockerized. If omitted, the best option will be selected automatically.
    rescan : {bool}
        True to force a rescan and replace existing metadata.
    tap_exe : {str}
        Overrides the tap executable, if different from `tap-{tap_name}`.
    target_exe : {str}
        Overrides the target executable, if different from `target-{tap_name}`.
    taps_dir : {str}
        The directory containing the rules file. (Default=cwd)
    config_dir : {str}
        The default location of config, catalog and other
        potentially sensitive information. (Recommended to be excluded from source control.)
        (Default="${taps_dir}/.secrets")
    config_file : {str}
        The location of the JSON config file which
        contains config for the specified tap or 'False' to only pull settings from
        environment variables. Default path is f"${config_dir}/${plugin_name}-config.json".
    catalog_dir : {str}
        The output directory to be used for saving catalog
        files. If not provided, a path will be generated automatically within `.output` or
        a path specified by the `TAP_SCRATCH_DIR` environment variable.
    target_config_file : {str}
        The location of the JSON config file which
        contains config for the specified target or 'False' to only pull settings from
        environment variables. Default path is f"${config_dir}/${plugin_name}-config.json".
    state_file : {str}
        The path to a state file. If not provided, a state
        file path will be generated automatically within `catalog_dir`.
    exclude_tables: {List(str)}
        A list of tables to exclude. Ignored
        if table_name arg is not "*".
    """
    config.print_version()

    tap_env_conf = config.get_plugin_settings_from_env(f"tap-{tap_name}")
    config_file = tap_env_conf.get("CONFIG_FILE", config_file)
    tap_exe = tap_exe or tap_env_conf.get("EXE", f"tap-{tap_name}")

    table_name = table_name or tap_env_conf.get("TABLE_NAME", None)
    exclude_tables = exclude_tables or tap_env_conf.get("EXCLUDE_TABLES", None)

    target_env_conf = config.get_plugin_settings_from_env(f"target-{target_name}")
    target_config_file = target_config_file or target_env_conf.get("CONFIG_FILE", None)
    target_exe = target_exe or target_env_conf.get("EXE", f"target-{target_name}")

    if dockerized is None:
        if uio.is_windows() or uio.is_mac():
            dockerized = True
            logging.info(
                "The 'dockerized' argument is not set when running either Windows or OSX..."
                "Defaulting to dockerized=True"
            )
        else:
            dockerized = False
    taps_dir = config.get_taps_dir(taps_dir)
    rules_file = config.get_rules_file(taps_dir, tap_name)
    config_required = True
    target_config_required = True
    logging.info(
        "Attempting to configure sync using config_file={config_file} and "
        "target_config_file={target_config_file}."
    )
    if (config_file is not None) and str(config_file).lower() == "false":
        logging.info("Skipping check for tap config (--config_file=False)")
        config_file = None
        config_required = False
    if (target_config_file is not None) and str(target_config_file).lower() == "false":
        logging.info("Skipping check for target config (--target_config_file=False)")
        target_config_file = None
        target_config_required = False
    config_file = str(
        config.get_config_file(
            f"tap-{tap_name}",
            taps_dir=taps_dir,
            config_dir=config_dir,
            config_file=config_file,
            required=config_required,
        )
    )
    target_config_file = str(
        config.get_config_file(
            f"target-{target_name}",
            taps_dir=taps_dir,
            config_dir=config_dir,
            config_file=target_config_file,
            required=target_config_required,
        )
    )
    if not uio.file_exists(config_file):
        raise FileExistsError(config_file)
    if not uio.file_exists(target_config_file):
        raise FileExistsError(target_config_file)
    catalog_dir = catalog_dir or config.get_catalog_output_dir(tap_name, taps_dir)
    full_catalog_file = f"{catalog_dir}/{tap_name}-catalog-selected.json"
    if rescan or rules_file or not uio.file_exists(full_catalog_file):
        plans.plan(
            tap_name,
            taps_dir=taps_dir,
            config_file=config_file,
            config_dir=catalog_dir,
            rescan=rescan,
            tap_exe=tap_exe,
        )
    list_of_tables = plans.get_table_list(
        table_filter=table_name,
        exclude_tables=exclude_tables,
        catalog_file=full_catalog_file,
    )
    logging.info(f"Table(s) to sync: {', '.join(list_of_tables)}")
    for table in list_of_tables:
        # Call each tap independently so that table state files are kept separate:
        tmp_catalog_file = f"{catalog_dir}/{tap_name}-{table}-catalog.json"
        table_state_file = (
            state_file
            or config.get_state_file_path()
            or f"{catalog_dir}/{table}-state.json"
        )
        plans._create_single_table_catalog(
            tap_name=tap_name,
            table_name=table,
            full_catalog_file=full_catalog_file,
            output_file=tmp_catalog_file,
        )
        _sync_one_table(
            tap_name=tap_name,
            target_name=target_name,
            table_name=table,
            config_file=config_file,
            target_config_file=target_config_file,
            table_catalog_file=tmp_catalog_file,
            table_state_file=table_state_file,
            dockerized=dockerized,
            tap_exe=tap_exe,
            target_exe=target_exe,
        )


def _dockerize_path(localpath: str, container_volume_root="/home/local") -> str:
    result = os.path.relpath(localpath).replace("\\", "/")
    result = f"{container_volume_root}/{result}"
    return result


def _dockerize_cli_args(arg_str: str, container_volume_root="/home/local") -> str:
    args = arg_str.split(" ")
    newargs: List[str] = []
    for arg in args:
        if uio.file_exists(arg):
            newargs.append(_dockerize_path(arg, container_volume_root))
        elif "=" in arg:
            left, right = arg.split("=")[0], "=".join(arg.split("=")[1:])
            if uio.file_exists(right):
                newargs.append(
                    f"{left}={_dockerize_path(right, container_volume_root)}"
                )
        else:
            newargs.append(arg)
    return " ".join(newargs)


@logged("running sync of table '{table_name}' from '{tap_name}'")
def _sync_one_table(
    tap_name: str,
    table_name: str,
    config_file: str,
    target_name: str,
    target_config_file: str,
    table_catalog_file: str,
    table_state_file: str,
    dockerized: bool,
    tap_exe: str,
    target_exe: str,
) -> None:
    if not tap_exe:
        tap_exe = f"tap-{tap_name}"
    pipeline_version_num = config.get_pipeline_version_number()
    table_state_file = config.replace_placeholders(
        {"table_state_file": table_state_file},
        tap_name,
        table_name,
        pipeline_version_num,
    )["table_state_file"]
    tap_args = f"--config {config_file} --catalog {table_catalog_file}"
    if uio.file_exists(table_state_file):
        local_state_file_in = os.path.join(
            uio.get_temp_dir(), f"{tap_name}-{table_name}-state.json"
        )
        if not uio.get_text_file_contents(table_state_file):
            logging.warning(f"Ignoring blank state file from '{table_state_file}'.")
        else:
            states.make_aggregate_state_file(table_state_file, local_state_file_in)
            tap_args += f" --state {local_state_file_in}"
        local_state_file_out = (
            f"{'.'.join(local_state_file_in.split('.')[:-1])}-new.json"
        )
    else:
        local_state_file_out = os.path.join(
            uio.get_temp_dir(), f"{tap_name}-{table_name}-state-new.json"
        )

    tmp_target_config = config.get_single_table_target_config_file(
        target_name,
        target_config_file,
        tap_name=tap_name,
        table_name=table_name,
        pipeline_version_num=pipeline_version_num,
    )
    target_args = f"--config {tmp_target_config}"
    if dockerized:
        cdw = os.getcwd().replace("\\", "/")
        tap_image_name = docker._get_docker_tap_image(tap_exe)
        target_image_name = docker._get_docker_tap_image(target_exe=target_exe)
        sync_cmd = (
            f"docker run --rm -it -v {cdw}:/home/local {tap_image_name} "
            f"{_dockerize_cli_args(tap_args)} "
            "| "
            f"docker run --rm -it -v {cdw}:/home/local {target_image_name} "
            f"{_dockerize_cli_args(target_args)} "
            ">> "
            f"{local_state_file_out}"
        )
    else:
        sync_cmd = (
            f"{tap_exe} "
            f"{tap_args} "
            "| "
            f"{target_exe} "
            f"{target_args} "
            "> "
            f"{local_state_file_out}"
        )
    runnow.run(sync_cmd)
    if not uio.file_exists(local_state_file_out):
        logging.warning(
            f"State file does not exist at path '{local_state_file_out}'. Skipping upload. "
            f"This can be caused by having no data, or no new data, in the source table."
        )
    else:
        uio.upload_file(local_state_file_out, table_state_file)



"""tapdance.docker - Module for containerization and image build functions."""

from typing import List, Optional, Tuple

# import dock_r
from logless import get_logger
import runnow
import uio
import yaml

from tapdance.config import SINGER_PLUGINS_INDEX

logging = get_logger("tapdance")

try:
    import dock_r
except Exception as ex:
    dock_r = None  # type: ignore
    logging.warning(f"Docker libraries were not able to be loaded ({ex}).")


BASE_DOCKER_REPO = "dataopstk/tapdance"


# Deprecated in favor of partial dockerization:
# def rerun_dockerized(
#     tap_alias: str,
#     target_alias: str = None,
#     *,
#     tap_exe: str,
#     target_exe: str,
#     args: List[str] = None,
# ) -> bool:
#     """Rerun the command from within docker.

#     Parameters
#     ----------
#     tap_alias : str
#         The name of the tap to use, without the tap- prefix.
#     target_alias : str, optional
#         Optional. The name of the target to use, without the tap- prefix.
#     args : List[str], optional
#         Optional. A list of command-line arguments. If omitted, will be parsed from the
#         command line.

#     Returns
#     -------
#     bool
#         True unless an error is raised.
#     """
#     if args is not None:
#         argstr = " ".join(args)
#     else:
#         argstr = " ".join(sys.argv[1:])
#     cmd = f"tapdance {argstr}"
#     env = {
#         k: v
#         for k, v in os.environ.items()
#         if (k.startswith("TAP_") or k.startswith("TARGET_"))
#     }
#     image_name = _get_docker_tap_image(tap_exe, target_exe)
#     try:
#         dock_r.pull(image_name)
#     except Exception as ex:
#         logging.warning(f"Could not pull docker image '{image_name}'. {ex}")
#     with logged_block(f"running dockerized command '{cmd}' on image '{image_name}'"):

#         def _build_docker_run(image, command, environment, working_dir, volumes):
#             e_str = " ".join([f"-e {k}={v}" for k, v in environment.items()])
#             w_str = f"-w {working_dir}" if working_dir else ""
#             v_str = " ".join([f"-v {x}:{y}" for x, y in volumes.items()])
#             docker_run_cmd = f"docker run {e_str} {v_str} {w_str} {image} {command}"
#             return docker_run_cmd

#         volumes = {os.path.abspath("."): "/projects/my-project"}
#         # DEV_DEBUG = False  # Used to test while developing, override python lib path
#         # if DEV_DEBUG:
#         #     container_lib = "/usr/local/lib/python3.8/site-packages/tapdance"
#         #     host_lib = "C:\\Files\\Source\\tapdance\\tapdance"
#         #     volumes[host_lib] = container_lib
#         docker_run_cmd = _build_docker_run(
#             image=image_name,
#             command=cmd,
#             environment=env,
#             working_dir="/projects/my-project",
#             volumes=volumes,
#         )
#         runnow.run(docker_run_cmd)
#     return True


def _get_docker_tap_image(
    tap_exe: Optional[str] = None, target_exe: Optional[str] = None
) -> str:
    if not tap_exe and not target_exe:
        raise ValueError("At least one value required of: tap_exe, target_exe")
    if target_exe:
        target_alias = target_exe.replace("target-", "")
    if tap_exe:
        tap_alias = tap_exe.replace("tap-", "")
        if target_exe:
            return f"{BASE_DOCKER_REPO}:{tap_alias}-to-{target_alias}"
        return f"{BASE_DOCKER_REPO}:tap-{tap_alias}"
    return f"{BASE_DOCKER_REPO}:target-{target_alias}"


def _get_plugins_list(
    plugins_index: Optional[str] = None,
) -> List[Tuple[str, str, str]]:
    plugins_index = plugins_index or SINGER_PLUGINS_INDEX
    if not uio.file_exists(plugins_index):
        raise RuntimeError(
            f"No file found at '{plugins_index}'."
            "Please set SINGER_PLUGINS_INDEX and try again."
        )
    yml_doc = yaml.safe_load(uio.get_text_file_contents(plugins_index))
    taps = yml_doc["singer-taps"]
    list_of_tuples = []
    taps = yml_doc["singer-taps"]
    targets = yml_doc["singer-targets"]
    plugins = taps + targets
    for plugin in plugins:
        list_of_tuples.append(
            (plugin["name"], plugin.get("source", None), plugin.get("alias", None),)
        )
    return list_of_tuples


def _build_all_standalone(
    source_image: Optional[str] = None,
    plugins_index: Optional[str] = None,
    push: bool = False,
    pre: bool = False,
    ignore_cache: bool = False,
) -> List[str]:
    """Build all standalone docker images.

    Parameters
    ----------
    source_image : str, optional
        Optional. Overrides the default base image.
    plugins_index : str, optional
        Optional. Overrides the path to the plugins index yaml file.
    push : bool, optional
        True to push the image after building, by default False.
    pre : bool, optional
        True to use the latest prerelease version, by default False.
    ignore_cache : bool, optional
        True to ignore rebuild all image steps, ignoring cache, by default False.

    Returns
    -------
    List[str]
        The list of images built.
    """
    plugins = _get_plugins_list(plugins_index)
    created_images = []
    for name, source, alias in plugins:
        created_images.append(
            _build_plugin_image(
                name,
                source=source,
                alias=alias,
                source_image=source_image,
                push=push,
                pre=pre,
                ignore_cache=ignore_cache,
            )
        )
    return created_images


def _get_plugin_info(
    plugin_id: str, plugins_index: Optional[str] = None
) -> Tuple[str, str, str]:
    plugins = _get_plugins_list(plugins_index)
    for name, source, alias in plugins:
        if (alias or name) == plugin_id:
            return (name, source, alias)
    raise ValueError(f"Could not locate the plugin called '{plugin_id}'")


def _build_all_composite(
    source_image: Optional[str] = None,
    plugins_index: Optional[str] = None,
    push: bool = False,
    pre: bool = False,
    ignore_cache: bool = False,
) -> List[str]:
    plugins = _get_plugins_list(plugins_index)
    created_images = []
    for tap_name, tap_source, tap_alias in plugins:
        tap_alias = tap_alias or tap_name
        for target_name, target_source, target_alias in plugins:
            target_alias = target_alias or target_name
            if tap_alias.startswith("tap-") and target_alias.startswith("target-"):
                has_custom_tap = (tap_source is not None) and (
                    "Dockerfile" in tap_source
                )
                has_custom_target = (target_source is not None) and (
                    "Dockerfile" in target_source
                )
                if has_custom_tap and has_custom_target:
                    logging.warning(
                        f"The '{tap_name}' tap is not compatible with "
                        f"'{target_name}' target because both require custom Dockerfiles. "
                        "Skipping..."
                    )
                else:
                    created_images.append(
                        _build_composite_image(
                            tap_alias,
                            target_alias,
                            push=push,
                            pre=pre,
                            ignore_cache=ignore_cache,
                            has_custom_tap=has_custom_tap,
                            has_custom_target=has_custom_target,
                        )
                    )
    return created_images


def _build_plugin_image(
    plugin_name: str,
    source: str,
    alias: str,
    source_image=None,
    push=False,
    pre=False,
    ignore_cache=False,
) -> str:
    source = source or plugin_name
    alias = alias or plugin_name
    image_name = f"{BASE_DOCKER_REPO}:{alias}"
    build_cmd = "docker build"
    if ignore_cache:
        build_cmd += " --no-cache"
    if source_image:
        build_cmd += f" --build-arg source_image={source_image}"
    if pre:
        build_cmd += " --build-arg prerelease=true"
        image_name += "--pre"
    if "Dockerfile" in source:
        dockerfile = source
    else:
        dockerfile = "singer-plugin.Dockerfile"
        build_cmd += f" --build-arg PLUGIN_SOURCE={source}"

    build_cmd += (
        f" --build-arg PLUGIN_NAME={plugin_name}"
        f" --build-arg PLUGIN_ALIAS={alias}"
        f" -t {image_name}"
        f" -f {dockerfile}"
        f" ."
    )
    runnow.run(build_cmd)
    if push:
        _push(image_name)
    return image_name


def _build_composite_image(
    tap_alias: str,
    target_alias: str,
    *,
    push: bool = False,
    pre: bool = False,
    ignore_cache: bool = False,
    has_custom_tap: bool,
    has_custom_target: bool,
) -> str:
    if has_custom_tap and has_custom_target:
        raise NotImplementedError(
            "Cannot combine a custom tap ('tap-{tap_alias}') "
            "with a custom target '{target_alias}'."
        )
    if tap_alias.startswith("tap-"):
        tap_alias = tap_alias.replace("tap-", "", 1)
    if target_alias.startswith("target-"):
        target_alias = target_alias.replace("target-", "", 1)
    image_name = f"{BASE_DOCKER_REPO}:{tap_alias}-to-{target_alias}"
    build_cmd = "docker build"
    if has_custom_tap:
        dockerfile = "tap-to-target-w-custom-tap.Dockerfile"
    else:
        dockerfile = "tap-to-target.Dockerfile"
    if ignore_cache:
        build_cmd += " --no-cache"
    if pre:
        build_cmd += " --build-arg source_image_suffix=--pre"
        image_name += "--pre"
    build_cmd += (
        f" --build-arg tap_alias={tap_alias}"
        f" --build-arg target_alias={target_alias}"
        f" -t {image_name}"
        f" -f {dockerfile}"
        f" ."
    )
    runnow.run(build_cmd)
    if push:
        _push(image_name)
    return image_name


def _push(image_name) -> None:
    runnow.run(f"docker push {image_name}")


def build_image(
    tap_or_plugin_alias: str,
    target_alias: str = None,
    push: bool = False,
    pre: bool = False,
    ignore_cache: bool = False,
) -> None:
    """Build a single image.

    If tap and target are both provided, any required upstream images will be built as well.

    Arguments:
        tap_or_plugin_alias {str} -- The name of the tap (without the `tap-` prefix).

    Keyword Arguments:
        target_alias {str} -- Optional. The name of the target (without the `target-` prefix).
        push {bool} -- True to push images to image repository after build. (default: {False})
        pre {bool} -- True to use and create prelease versions. (default: {False})
        ignore_cache {bool} -- True to build images without cached image layers. (default: {False})
    """
    has_custom_tap = uio.file_exists(f"./tap-{tap_or_plugin_alias}.Dockerfile")
    has_custom_target = uio.file_exists(f"./target-{target_alias}.Dockerfile")
    if has_custom_tap and has_custom_target:
        raise NotImplementedError(
            "Cannot combine a custom tap ('tap-{tap_or_plugin_alias}') "
            "with a custom target '{target_alias}'."
        )
    if has_custom_tap:
        logging.info(f"Using custom Dockerfile for tap-{tap_or_plugin_alias}")
        _build_plugin_image(
            f"tap-{tap_or_plugin_alias}",
            source=f"./tap-{tap_or_plugin_alias}.Dockerfile",
            alias=f"tap-{tap_or_plugin_alias}",
            push=push,
            pre=pre,
            ignore_cache=ignore_cache,
        )
    else:
        name, source, alias = _get_plugin_info(f"tap-{tap_or_plugin_alias}")
        if source and "Dockerfile" in source:
            has_custom_tap = True
        _build_plugin_image(
            name,
            source=source,
            alias=alias,
            push=push,
            pre=pre,
            ignore_cache=ignore_cache,
        )
    if target_alias:
        if has_custom_target:
            logging.info(f"Using custom Dockerfile for target-{target_alias}")
            _build_plugin_image(
                f"target-{target_alias}",
                source=f"./target-{target_alias}.Dockerfile",
                alias=f"target-{target_alias}",
                push=push,
                pre=pre,
                ignore_cache=ignore_cache,
            )
        else:
            name, source, alias = _get_plugin_info(f"target-{target_alias}")
            if "Dockerfile" in source:
                has_custom_target = True
            _build_plugin_image(
                name,
                source=source,
                alias=alias,
                push=push,
                pre=pre,
                ignore_cache=ignore_cache,
            )
        _build_composite_image(
            tap_alias=tap_or_plugin_alias,
            target_alias=target_alias,
            push=push,
            pre=pre,
            ignore_cache=ignore_cache,
            has_custom_tap=has_custom_tap,
            has_custom_target=has_custom_target,
        )


def build_all_images(
    push: bool = False, pre: bool = False, ignore_cache: bool = False
) -> List[str]:
    """
    Build all images.

    :param push: Push images after building
    :param pre: Create and publish pre-release builds
    :param ignore_cache: True to build images without cached image layers. (default: {False})
    """
    built_images = _build_all_standalone(push=push, pre=pre, ignore_cache=ignore_cache)
    built_images += _build_all_composite(push=push, pre=pre, ignore_cache=ignore_cache)
    return built_images



"""tapdance.states - Defines state aggregation and state handling helper functions."""

import json

from logless import get_logger
import uio

logging = get_logger("tapdance")


def _is_valid_json(json_text: str) -> bool:
    try:
        _ = json.loads(json_text)
    except ValueError:
        return False
    return True


def make_aggregate_state_file(raw_json_lines_file: str, output_json_file: str) -> None:
    """
    Create a valid json state file from one or more json lines ('jsonl' format).

    Parameters
    ----------
    raw_json_lines_file : str
        Path to a jsonl (json lines) file containing one or more json documents to
        aggregate.
    output_json_file : str
        Path to use when saving the aggregated json file.
    """
    try:
        uio.create_text_file(
            output_json_file,
            get_aggregate_state(uio.get_text_file_contents(raw_json_lines_file)),
        )
    except ValueError as ex:
        raise ValueError(
            f"State file from '{raw_json_lines_file}' is not valid JSON or JSONL. "
            f"Please either delete or fix the file and then retry. {ex}"
        )


def get_aggregate_state(raw_json_lines_text: str) -> str:
    """
    Return a valid aggregated json state string from one or json lines ('jsonl' format).

    Parameters
    ----------
    raw_json_lines_text : str
        String jsonl (json lines) with one or more json documents to aggregate.
    """
    if raw_json_lines_text == "":
        raise ValueError("Cannot aggregate json state - text is empty string.")
    if _is_valid_json(raw_json_lines_text):
        return raw_json_lines_text
    elif _is_valid_json(raw_json_lines_text.splitlines()[-1]):
        logging.warning(
            "State file contains multiple states. Using final line of JSON state: "
            + raw_json_lines_text.replace("\n", "\\n")
        )
        return raw_json_lines_text.splitlines()[-1]
    elif len(raw_json_lines_text.splitlines()) >= 2 and _is_valid_json(
        raw_json_lines_text.splitlines()[-2]
    ):
        logging.warning(
            "State file contains multiple states. "
            "Using 2nd-to-last line of JSON state: "
            + raw_json_lines_text.replace("\n", "\\n")
        )
        return raw_json_lines_text.splitlines()[-2]
    else:
        raise ValueError(
            "State is not valid JSON: " + raw_json_lines_text.replace("\n", "\\n")
        )
